"""CLI info command."""
from argparse import ArgumentParser, Namespace

from InquirerPy.utils import color_print  # type: ignore
from InquirerPy.separator import Separator  # type: ignore

from redbrick.cli.project import CLIProject
from redbrick.cli.cli_base import CLIInfoInterface


class CLIInfoController(CLIInfoInterface):
    """CLI info command controller."""

    def __init__(self, parser: ArgumentParser) -> None:
        """Intialize info sub commands."""
        parser.add_argument(
            "path",
            nargs="?",
            default=".",
            help="Path of project (Default: current directory)",
        )

    def handler(self, args: Namespace) -> None:
        """Handle info command."""
        self.args = args
        project = CLIProject.from_path(path=self.args.path)
        assert project, "Not a valid project"
        self.project = project

        self.handle_info()

    def handle_info(self) -> None:
        """Handle empty sub command."""
        org = self.project.org
        project = self.project.project

        org_data, project_data = [], []
        org_data.append(f"ID: {org.org_id}")
        org_data.append(f"Name: {org.name}")

        project_data.append(f"ID: {project.project_id}")
        project_data.append(f"Name: {project.name}")
        project_data.append(f"Type: {project.project_type.value}")
        project_data.append(f"Taxonomy: {project.taxonomy_name}")
        project_data.append(f"URL: {project.url}")

        color_print(
            [("green", f"{Separator()} Organization {Separator()}\n")]
            + [("", f"{item}\n") for item in org_data]
            + [("green", f"{Separator()} Project {Separator()}\n")]
            + [("", f"{item}\n") for item in project_data]
        )



"""
Configure pytest.

Reusable fixtures.
"""
import os

import pytest

from redbrick import _populate_context
from redbrick.common.context import RBContext
from redbrick.organization import RBOrganization
from redbrick.project import RBProject


@pytest.fixture(name="api_key")
def fixture_api_key() -> str:
    """Get api_key."""
    return os.environ.get("REDBRICK_SDK_API_KEY", "")


@pytest.fixture(name="url")
def fixture_url() -> str:
    """Get url."""
    return os.environ.get("REDBRICK_SDK_URL", "")


@pytest.fixture(name="org_id")
def fixture_org_id() -> str:
    """Get org_id."""
    return os.environ.get("REDBRICK_SDK_ORG_ID", "")


@pytest.fixture(name="project_id")
def fixture_project_id() -> str:
    """Get project_id."""
    return os.environ.get("REDBRICK_SDK_PROJECT_ID", "")


@pytest.fixture(name="context")
def fixture_context(api_key: str, url: str) -> RBContext:
    """Get RBContext."""
    return _populate_context(RBContext(api_key=api_key, url=url))


@pytest.fixture(name="org")
def fixture_org(context: RBContext, org_id: str) -> RBOrganization:
    """Get RBOrganization."""
    return RBOrganization(context, org_id)


@pytest.fixture(name="project")
def fixture_project(context: RBContext, org_id: str, project_id: str) -> RBProject:
    """Get RBProject."""
    return RBProject(context, org_id, project_id)



"""Partial queries to prevent duplication."""

TAXONOMY_SHARD = """
name
version
categories {
    name
    children {
        name
        classId
        children {
            name
            classId
            children {
                name
                classId
                children {
                    name
                    classId
                    children {
                        name
                        classId
                        children {
                            name
                            classId
                        }
                    }
                }
            }
        }
    }
}
"""

LATEST_TASKDATA_SHARD = """
    latestTaskData {
        dataPoint {
            name
            itemsPresigned: items(presigned: true)
            items(presigned: false)
        }
        createdByEmail
        labelsData(interpolate: true)
        labelsPath
    }
"""



"""Utilities for working with label objects."""

from typing import Dict, List, Optional


def clean_rb_label(label: Dict) -> Dict:
    """Clean any None fields."""
    for key, val in label.copy().items():
        if val is None:
            del label[key]
    return label


def flat_rb_format(
    labels: List[Dict],
    items: List[str],
    items_presigned: List[str],
    name: str,
    created_by: str,
    task_id: str,
    current_stage_name: str,
    labels_path: Optional[str],
) -> Dict:
    """Get standard rb flat format, same as import format."""
    return {
        "labels": labels,
        "items": items,
        "itemsPresigned": items_presigned,
        "name": name,
        "taskId": task_id,
        "createdBy": created_by,
        "currentStageName": current_stage_name,
        "labelsPath": labels_path,
    }



"""Public API to exporting."""
import asyncio
from typing import List, Dict, Optional, Tuple, Any
from functools import partial
import os
import json
import copy
from datetime import datetime, timezone

import aiohttp
from shapely.geometry import Polygon  # type: ignore
from skimage.morphology import remove_small_holes  # type: ignore
import numpy as np  # type: ignore
from matplotlib import cm  # type: ignore
import tqdm  # type: ignore
from PIL import Image  # type: ignore

from redbrick.common.constants import MAX_CONCURRENCY
from redbrick.common.context import RBContext
from redbrick.common.enums import LabelType
from redbrick.utils.async_utils import gather_with_concurrency
from redbrick.utils.files import uniquify_path, download_files
from redbrick.utils.logging import (
    print_error,
    print_info,
    print_warning,
    handle_exception,
)
from redbrick.utils.pagination import PaginationIterator
from redbrick.utils.rb_label_utils import clean_rb_label, flat_rb_format
from redbrick.coco.coco_main import coco_converter


def _parse_entry_latest(item: Dict) -> Dict:
    try:
        task_id = item["taskId"]
        task_data = item["latestTaskData"]
        datapoint = task_data["dataPoint"]
        items_presigned = datapoint["itemsPresigned"]
        items = datapoint["items"]
        name = datapoint["name"]
        created_by = task_data["createdByEmail"]
        labels = [
            clean_rb_label(label) for label in json.loads(task_data["labelsData"])
        ]

        return flat_rb_format(
            labels,
            items,
            items_presigned,
            name,
            created_by,
            task_id,
            item["currentStageName"],
            task_data["labelsPath"],
        )
    except (AttributeError, KeyError, TypeError, json.decoder.JSONDecodeError):
        return {}


class Export:
    """
    Primary interface to handling export from a project.

    This class has methods to export to various formats depending on
    your project type.
    """

    def __init__(
        self, context: RBContext, org_id: str, project_id: str, project_type: LabelType
    ) -> None:
        """Construct Export object."""
        self.context = context
        self.org_id = org_id
        self.project_id = project_id
        self.project_type = project_type

    def _get_raw_data_latest(
        self,
        concurrency: int,
        only_ground_truth: bool = False,
        from_timestamp: Optional[float] = None,
    ) -> Tuple[List[Dict], Dict]:
        temp = self.context.export.get_datapoints_latest
        stage_name = "END" if only_ground_truth else None
        my_iter = PaginationIterator(
            partial(
                temp,
                self.org_id,
                self.project_id,
                stage_name,
                datetime.fromtimestamp(from_timestamp, tz=timezone.utc)
                if from_timestamp is not None
                else None,
                concurrency,
            )
        )

        general_info = self.context.export.get_output_info(self.org_id, self.project_id)
        datapoint_count = self.context.export.datapoints_in_project(
            self.org_id, self.project_id, stage_name
        )

        print_info(
            f"Downloading {'groundtruth' if only_ground_truth else 'all'} tasks"
            + (
                f" updated since {datetime.fromtimestamp(from_timestamp)}"
                if from_timestamp is not None
                else ""
            )
        )

        with tqdm.tqdm(my_iter, unit=" datapoints", total=datapoint_count) as progress:
            datapoints = []
            for val in progress:
                task = _parse_entry_latest(val)
                if task:
                    datapoints.append(task)
            disable = progress.disable
            progress.disable = False
            progress.update(datapoint_count - progress.n)
            progress.disable = disable

        return datapoints, general_info["taxonomy"]

    async def _get_input_labels(self, dp_ids: List[str]) -> List[Dict]:
        conn = aiohttp.TCPConnector(limit=MAX_CONCURRENCY)
        async with aiohttp.ClientSession(connector=conn) as session:
            coros = [
                self.context.export.get_labels(
                    session, self.org_id, self.project_id, dp_id
                )
                for dp_id in dp_ids
            ]
            input_data = await gather_with_concurrency(MAX_CONCURRENCY, coros)

        await asyncio.sleep(0.250)  # give time to close ssl connections
        labels = [
            [clean_rb_label(label) for label in json.loads(data["labelsData"])]
            for data in input_data
        ]
        return [
            {**data, "inputLabels": label} for data, label in zip(input_data, labels)
        ]

    def get_latest_data(
        self,
        concurrency: int,
        only_ground_truth: bool = False,
        from_timestamp: Optional[float] = None,
        get_input_labels: bool = False,
    ) -> Tuple[List[Dict], float]:
        """Export latest data."""
        # pylint: disable=too-many-locals
        stage_name = "END" if only_ground_truth else None
        cache_time = (
            datetime.fromtimestamp(from_timestamp, tz=timezone.utc)
            if from_timestamp is not None
            else None
        )
        print_info(
            f"Downloading {'groundtruth' if only_ground_truth else 'all'} tasks"
            + (
                f" updated since {datetime.fromtimestamp(from_timestamp)}"
                if from_timestamp is not None
                else ""
            )
        )
        tasks = []
        cursor = None
        new_cache_time = cache_time
        dp_ids = []
        while True:
            entries, cursor, new_cache_time = self.context.export.get_datapoints_latest(
                self.org_id,
                self.project_id,
                stage_name,
                cache_time,
                concurrency,
                cursor,
            )
            for val in entries:
                tasks.append(_parse_entry_latest(val))
                dp_ids.append(val["dpId"])

            if cursor is None:
                break

        if get_input_labels:
            loop = asyncio.get_event_loop()
            input_labels = loop.run_until_complete(self._get_input_labels(dp_ids))
            for idx, input_label in enumerate(input_labels):
                tasks[idx] = {**tasks[idx], "inputLabels": input_label["inputLabels"]}

        return tasks, (new_cache_time.timestamp() if new_cache_time else 0)

    def _get_raw_data_single(self, task_id: str) -> Tuple[List[Dict], Dict]:
        general_info = self.context.export.get_output_info(self.org_id, self.project_id)
        datapoint = self.context.export.get_datapoint_latest(
            self.org_id, self.project_id, task_id
        )
        datapoints = []
        task = _parse_entry_latest(datapoint)
        if task:
            datapoints.append(task)
        return datapoints, general_info["taxonomy"]

    @staticmethod
    def _get_color(class_id: int) -> Any:
        """Get a color from class id."""
        if class_id > 20:
            color = (
                np.array(cm.tab20b(int(class_id))) * 255  # pylint: disable=no-member
            )
            return color.astype(np.uint8)

        color = np.array(cm.tab20c(int(class_id))) * 255  # pylint: disable=no-member
        return color.astype(np.uint8)

    @staticmethod
    def tax_class_id_mapping(
        taxonomy: Dict, class_id: Dict, color_map: Optional[Dict] = None
    ) -> None:
        """Create a class mapping from taxonomy categories to class_id."""
        for category in taxonomy:
            class_id[category["name"]] = category["classId"] + 1

            # Create a color map
            if color_map is not None:
                color_map[category["name"]] = Export._get_color(category["classId"])[
                    0:3
                ].tolist()  # not doing +1 here.

            Export.tax_class_id_mapping(category["children"], class_id, color_map)

    @staticmethod
    def fill_mask_holes(mask: np.ndarray, max_hole_size: int) -> np.ndarray:
        """Fill holes."""
        mask_copy = copy.deepcopy(mask)

        # find indexes where mask has labels
        mask_greater_zero = np.where(mask > 0)

        # convery copy mask to binary
        mask_copy[mask_greater_zero] = 1

        # fill holes in copy binary mask
        mask_copy = remove_small_holes(
            mask_copy.astype(bool),
            area_threshold=max_hole_size,
        )
        mask_copy = mask_copy.astype(int)

        # set original pixel values
        mask_copy[mask_greater_zero] = mask[mask_greater_zero]

        # find indexes of holes, and fill with neighbor
        mask_hole_loc = np.where((mask == 0) & (mask_copy > 0))

        for i in range(len(mask_hole_loc[0])):
            mask_copy = Export.fill_hole_with_neighbor(
                mask_copy, mask_hole_loc[0][i], mask_hole_loc[1][i]
            )

        return mask_copy

    @staticmethod
    def fill_hole_with_neighbor(mask: np.ndarray, i: Any, j: Any) -> np.ndarray:
        """Fill a pixel in the mask with it's neighbors value."""
        row, col = mask.shape
        top = 0 if j - 1 < 0 else mask[i][j - 1]
        top_right = 0 if (j - 1 < 0) or (i + 1 == row) else mask[i + 1][j - 1]
        right = 0 if i + 1 == row else mask[i + 1][j]
        bottom_right = 0 if (j + 1 == col) or (i + 1 == row) else mask[i + 1][j + 1]
        bottom = 0 if j + 1 == col else mask[i][j + 1]
        bottom_left = 0 if (i - 1 < 0) or (j + 1 == col) else mask[i - 1][j + 1]
        left = 0 if i - 1 < 0 else mask[i - 1][j]
        top_left = 0 if (i - 1 < 0) or (j - 1 == 0) else mask[i - 1][j - 1]
        mask[i][j] = max(
            top,
            top_right,
            right,
            bottom_right,
            bottom,
            bottom_left,
            left,
            top_left,
        )
        return mask

    @staticmethod
    def convert_rbai_mask(  # pylint: disable=too-many-locals
        labels: List,
        class_id_map: Dict,
        fill_holes: bool = False,
        max_hole_size: int = 30,
    ) -> np.ndarray:
        """Convert rbai datapoint to a numpy mask."""
        try:
            import rasterio.features  # pylint: disable=import-error, import-outside-toplevel
        except Exception as error:
            print_error(
                "For windows users, please follow the rasterio "
                + "documentation to properly install the module "
                + "https://rasterio.readthedocs.io/en/latest/installation.html "
                + "Rasterio is required by RedBrick SDK to work with masks."
            )
            raise error

        imagesize = labels[0]["pixel"]["imagesize"]

        # deal with condition where imagesize is returned as float
        imagesize = np.round(imagesize).astype(int)  # type: ignore

        mask = np.zeros([imagesize[1], imagesize[0]])

        for label in labels:
            class_id = class_id_map[label["category"][0][-1]]
            regions = copy.deepcopy(label["pixel"]["regions"])
            holes = copy.deepcopy(label["pixel"]["holes"])
            imagesize = label["pixel"]["imagesize"]

            # deal with condition where imagesize is returned as float
            imagesize = np.round(imagesize).astype(int)  # type: ignore

            # iterate through regions, and create region mask
            region_mask = np.zeros([imagesize[1], imagesize[0]])
            if regions and len(regions) > 0:
                for region in regions:
                    if (
                        len(np.array(region).shape) == 1
                        or np.array(region).shape[0] < 3
                    ):
                        # Don't add empty regions to the mask
                        # Don't add regions with < 3 vertices
                        break

                    # convert polygon to mask
                    region_polygon = Polygon(region)
                    single_region_mask = (
                        rasterio.features.rasterize(
                            [region_polygon],
                            out_shape=(imagesize[1], imagesize[0]),
                        ).astype(float)
                        * class_id
                    )

                    # add single region to root region mask
                    region_mask += single_region_mask

            # iterate through holes, and create hole mask
            hole_mask = np.zeros([imagesize[1], imagesize[0]])
            if holes and len(holes) > 0:
                for hole in holes:
                    if len(np.array(hole).shape) == 1 or np.array(hole).shape[0] < 3:
                        # Don't add empty hole to negative mask
                        # Don't add holes with < 3 vertices
                        break

                    # convert polygon hole to mask
                    hole_polygon = Polygon(hole)
                    single_hole_mask = (
                        rasterio.features.rasterize(
                            [hole_polygon],
                            out_shape=(imagesize[1], imagesize[0]),
                        ).astype(float)
                        * class_id
                    )

                    # add single hole mask to total hole mask
                    hole_mask += single_hole_mask

            # subtract the hole mask from region mask
            region_mask -= hole_mask

            # cleanup:
            # - remove overlapping region values
            neg_idxs = np.where(region_mask < 0)
            region_mask[neg_idxs] = 100
            # - remove negative values from overlapping holes
            overlap_indexes = np.where(region_mask > class_id)
            region_mask[overlap_indexes] = 100

            # merge current object to main mask
            class_idx_not_zero = np.where(region_mask != 0)
            mask[class_idx_not_zero] = region_mask[class_idx_not_zero]

            # fill all single pixel holes
            if fill_holes:
                mask = Export.fill_mask_holes(mask, max_hole_size)

            # convert 2d mask into 3d mask with colors
            color_mask = np.zeros((mask.shape[0], mask.shape[1], 3))
            class_ids = np.unique(mask)  # type: ignore
            for i in class_ids:
                if i == 0:
                    # don't add color to background
                    continue
                indexes = np.where(mask == i)
                color_mask[indexes] = Export._get_color(i - 1)[0:3]

        return color_mask

    @staticmethod
    def _export_png_mask_data(
        datapoints: List[Dict],
        taxonomy: Dict,
        mask_dir: str,
        class_map: str,
        datapoint_map: str,
        fill_holes: bool = False,
        max_hole_size: int = 30,
    ) -> None:
        """Export png masks and map json."""
        # pylint: disable=too-many-locals
        # Create a color map from the taxonomy
        class_id_map: Dict = {}
        color_map: Dict = {}
        Export.tax_class_id_mapping(
            taxonomy["categories"][0]["children"], class_id_map, color_map
        )

        # Convert rbai to png masks and save output
        dp_map = {}
        print_info("Converting to masks")

        for datapoint in tqdm.tqdm(datapoints):
            labels = [label for label in datapoint["labels"] if "pixel" in label]
            if not labels:
                print_warning(
                    f"No segmentation labels in task {datapoint['taskId']}, skipping"
                )
                continue

            filename = f"{datapoint['taskId']}.png"
            dp_map[filename] = datapoint["items"][0]

            color_mask = Export.convert_rbai_mask(
                labels, class_id_map, fill_holes, max_hole_size
            )

            # save png as 3 channel np.uint8 image
            pil_color_mask = Image.fromarray(color_mask.astype(np.uint8))
            pil_color_mask.save(os.path.join(mask_dir, filename))

        with open(class_map, "w", encoding="utf-8") as file_:
            json.dump(color_map, file_, indent=2)

        with open(datapoint_map, "w", encoding="utf-8") as file_:
            json.dump(dp_map, file_, indent=2)

    @handle_exception
    def redbrick_png(  # pylint: disable=too-many-locals
        self,
        only_ground_truth: bool = True,
        concurrency: int = 10,
        task_id: Optional[str] = None,
        fill_holes: bool = False,
        max_hole_size: int = 30,
        from_timestamp: Optional[float] = None,
    ) -> None:
        """
        Export segmentation labels as masks.

        Masks are exported to a local directory named after project_id.
        Please visit https://docs.redbrickai.com/python-sdk/reference#png-mask-formats
        to see an overview of the format of the exported masks.

        >>> project = redbrick.get_project(api_key, url, org_id, project_id)
        >>> project.export.redbrick_png()

        Parameters
        --------------
        only_ground_truth: bool = True
            If set to True, will only return data that has
            been completed in your workflow. If False, will
            export latest state

        concurrency: int = 10

        task_id: Optional[str] = None
            If the unique task_id is mentioned, only a single
            datapoint will be exported.

        fill_holes : bool = False
            If set to True, will fill any holes in your segmentation
            masks.

        max_hole_size: int = 10
            If fill_holes = True, this parameter defines the maximum
            size hole, in pixels, to fill.

        from_timestamp: Optional[float] = None
            If the timestamp is mentioned, will only export tasks
            that were labeled/updated since the given timestamp.

        Warnings
        ----------
        redbrick_png only works for the following types - IMAGE_SEGMENTATION, IMAGE_MULTI

        """
        if self.project_type not in (
            LabelType.IMAGE_SEGMENTATION,
            LabelType.IMAGE_MULTI,
        ):
            print_error(
                f"Project type needs to be {LabelType.IMAGE_SEGMENTATION} or "
                + f"{LabelType.IMAGE_MULTI} for redbrick_png"
            )
            return

        if task_id:
            datapoints, taxonomy = self._get_raw_data_single(task_id)
        else:
            datapoints, taxonomy = self._get_raw_data_latest(
                concurrency, only_ground_truth, from_timestamp
            )

        # Create output directory
        output_dir = uniquify_path(self.project_id)
        mask_dir = os.path.join(output_dir, "masks")
        os.makedirs(mask_dir, exist_ok=True)
        print_info(f"Saving masks to {output_dir} directory")

        Export._export_png_mask_data(
            datapoints,
            taxonomy,
            mask_dir,
            os.path.join(output_dir, "class_map.json"),
            os.path.join(output_dir, "datapoint_map.json"),
            fill_holes,
            max_hole_size,
        )

    @staticmethod
    def _export_nifti_label_data(
        datapoints: List[Dict], nifti_dir: str, task_map: str
    ) -> None:
        files = []
        for datapoint in datapoints:
            files.append(
                (
                    datapoint["labelsPath"],
                    os.path.join(nifti_dir, f"{datapoint['taskId']}.nii"),
                )
            )
        loop = asyncio.get_event_loop()
        paths: List[Optional[str]] = loop.run_until_complete(download_files(files))

        tasks = [
            {
                "labelsPath": path,
                **{
                    key: value
                    for key, value in datapoint.items()
                    if key not in ("currentStageName", "labelsPath")
                },
            }
            for datapoint, path in zip(datapoints, paths)
        ]

        with open(task_map, "w", encoding="utf-8") as tasks_file:
            json.dump(tasks, tasks_file, indent=2)

    @handle_exception
    def redbrick_nifti(
        self,
        only_ground_truth: bool = True,
        concurrency: int = 10,
        task_id: Optional[str] = None,
        from_timestamp: Optional[float] = None,
    ) -> None:
        """
        Export dicom segmentation labels in NIfTI-1 format.

        >>> project = redbrick.get_project(api_key, url, org_id, project_id)
        >>> project.export.redbrick_nifti()

        Parameters
        --------------
        only_ground_truth: bool = True
            If set to True, will only return data that has
            been completed in your workflow. If False, will
            export latest state

        concurrency: int = 10

        task_id: Optional[str] = None
            If the unique task_id is mentioned, only a single
            datapoint will be exported.

        from_timestamp: Optional[float] = None
            If the timestamp is mentioned, will only export tasks
            that were labeled/updated since the given timestamp.

        Warnings
        ----------
        redbrick_nifti only works for the following types - DICOM_SEGMENTATION

        """
        if self.project_type != LabelType.DICOM_SEGMENTATION:
            print_error(
                f"Project type needs to be {LabelType.DICOM_SEGMENTATION} "
                + "for redbrick_nifi"
            )
            return

        if task_id:
            datapoints, _ = self._get_raw_data_single(task_id)
        else:
            datapoints, _ = self._get_raw_data_latest(
                concurrency, only_ground_truth, from_timestamp
            )

        # Create output directory
        destination = uniquify_path(self.project_id)
        nifti_dir = os.path.join(destination, "nifti")
        os.makedirs(nifti_dir, exist_ok=True)
        print_info(f"Saving NIfTI files to {destination} directory")
        Export._export_nifti_label_data(
            datapoints, nifti_dir, os.path.join(destination, "tasks.json")
        )

    @handle_exception
    def redbrick_format(
        self,
        only_ground_truth: bool = True,
        concurrency: int = 10,
        task_id: Optional[str] = None,
        from_timestamp: Optional[float] = None,
    ) -> List[Dict]:
        """
        Export data into redbrick format.

        >>> project = redbrick.get_project(api_key, url, org_id, project_id)
        >>> result = project.export.redbrick_format()

        Parameters
        -----------------
        only_ground_truth: bool = True
            If set to True, will only return data that has
            been completed in your workflow. If False, will
            export latest state

        concurrency: int = 10

        task_id: Optional[str] = None
            If the unique task_id is mentioned, only a single
            datapoint will be exported.

        from_timestamp: Optional[float] = None
            If the timestamp is mentioned, will only export tasks
            that were labeled/updated since the given timestamp.

        Returns:
        -----------------
        List[Dict]
            Datapoint and labels in RedBrick AI format. See
            https://docs.redbrickai.com/python-sdk/reference
        """
        if task_id:
            datapoints, _ = self._get_raw_data_single(task_id)
        else:
            datapoints, _ = self._get_raw_data_latest(
                concurrency, only_ground_truth, from_timestamp
            )

        return [
            {
                key: value
                for key, value in datapoint.items()
                if key not in ("currentStageName", "labelsPath")
            }
            for datapoint in datapoints
        ]

    @handle_exception
    def coco_format(
        self,
        only_ground_truth: bool = True,
        concurrency: int = 10,
        task_id: Optional[str] = None,
        from_timestamp: Optional[float] = None,
    ) -> Dict:
        """
        Export project into coco format.

        >>> project = redbrick.get_project(api_key, url, org_id, project_id)
        >>> result = project.export.coco_format()

        Parameters
        -----------
        only_ground_truth: bool = True
            If set to True, will only return data that has
            been completed in your workflow. If False, will
            export latest state

        concurrency: int = 10

        task_id: Optional[str] = None
            If the unique task_id is mentioned, only a single
            datapoint will be exported.

        from_timestamp: Optional[float] = None
            If the timestamp is mentioned, will only export tasks
            that were labeled/updated since the given timestamp.

        Returns
        -----------
        List[Dict]
            Datapoint and labels in COCO format. See
            https://cocodataset.org/#format-data

        Warnings
        ----------
        redbrick_coco only works for the following types - IMAGE_BBOX, IMAGE_POLYGON
        """
        if self.project_type not in (LabelType.IMAGE_BBOX, LabelType.IMAGE_POLYGON):
            print_error(
                f"Project type needs to be {LabelType.IMAGE_BBOX} or "
                + f"{LabelType.IMAGE_POLYGON} for redbrick_coco"
            )
            return {}

        if task_id:
            datapoints, taxonomy = self._get_raw_data_single(task_id)
        else:
            datapoints, taxonomy = self._get_raw_data_latest(
                concurrency, only_ground_truth, from_timestamp
            )

        return coco_converter(datapoints, taxonomy)



"""CLI commands controllers."""
from redbrick.cli.command.config import CLIConfigController
from redbrick.cli.command.init import CLIInitController
from redbrick.cli.command.clone import CLICloneController
from redbrick.cli.command.info import CLIInfoController
from redbrick.cli.command.export import CLIExportController
from redbrick.cli.command.prune import CLIPruneController
from redbrick.cli.command.upload import CLIUploadController



"""Repo for accessing export apis."""
from typing import Optional, List, Dict, Tuple
from datetime import datetime
from dateutil import parser

import aiohttp

from redbrick.common.export import ExportControllerInterface
from redbrick.common.client import RBClient
from redbrick.repo.shards import TAXONOMY_SHARD, LATEST_TASKDATA_SHARD


class ExportRepo(ExportControllerInterface):
    """Handle API requests to get export data."""

    def __init__(self, client: RBClient) -> None:
        """Construct ExportRepo."""
        self.client = client

    def get_datapoints_input(
        self,
        org_id: str,
        project_id: str,
        first: int = 50,
        presign: bool = False,
        cursor: Optional[str] = None,
    ) -> Tuple[List[Dict], Optional[str]]:
        """Get datapoints that were uploaded to the project."""
        raise NotImplementedError()

    def get_output_info(self, org_id: str, project_id: str) -> Dict:
        """Get info about the output labelset and taxonomy."""
        query_string = (
            """
        query ($orgId: UUID!, $name: String!){
            customGroup(orgId: $orgId, name:$name){
                dataType
                taskType
                datapointCount
                taxonomy {
                    %s
                    colorMap {
                        name
                        color
                        classid
                        trail
                        taskcategory
                    }
                }
            }
        }
        """
            % TAXONOMY_SHARD
        )

        # EXECUTE THE QUERY
        query_variables = {
            "orgId": org_id,
            "name": project_id + "-output",
        }

        result = self.client.execute_query(query_string, query_variables)

        temp: Dict = result["customGroup"]
        return temp

    def get_datapoints_output(
        self,
        org_id: str,
        project_id: str,
        first: int = 50,
        cursor: Optional[str] = None,
    ) -> Tuple[List[Dict], Optional[str]]:
        """Get datapoints that have made it to the output of the project."""
        query_string = """
        query ($orgId: UUID!, $projectId: UUID!, $name: String!,$first: Int, $cursor: String){
        customGroup(orgId: $orgId, name:$name){
            datapointsPaged(first:$first, after:$cursor) {
            entries {
                name
                itemsPresigned:items (presigned:true)
                items(presigned:false)
                task(projectId: $projectId) {
                    taskId
                }
                labelData(customGroupName: $name){
                    createdByEmail
                    labelsData(interpolate: true)
                    labelsPath
                }
            }
            cursor
            }
        }
        }
        """
        # EXECUTE THE QUERY
        query_variables = {
            "orgId": org_id,
            "name": project_id + "-output",
            "projectId": project_id,
            "cursor": cursor,
            "first": first,
        }

        result = self.client.execute_query(query_string, query_variables, False)
        custom_group = result.get("customGroup", {}) or {}
        datapoints_paged = custom_group.get("datapointsPaged", {}) or {}
        entries: List[Dict] = datapoints_paged.get("entries", []) or []  # type: ignore
        return entries, datapoints_paged.get("cursor")

    def datapoints_in_project(
        self, org_id: str, project_id: str, stage_name: Optional[str] = None
    ) -> int:
        """Get number of datapoints in project."""
        query_string = """
        query($orgId: UUID!, $projectId: UUID!, $stageName: String, $first: Int) {
            tasksPaged(
                orgId: $orgId
                projectId: $projectId
                stageName: $stageName
                first: $first
            ) {
                count
            }
        }
        """
        # EXECUTE THE QUERY
        query_variables = {
            "orgId": org_id,
            "projectId": project_id,
            "stageName": stage_name,
            "first": 0,
        }

        result = self.client.execute_query(query_string, query_variables)

        return int(result["tasksPaged"]["count"])

    def get_datapoints_latest(
        self,
        org_id: str,
        project_id: str,
        stage_name: Optional[str] = None,
        cache_time: Optional[datetime] = None,
        first: int = 50,
        cursor: Optional[str] = None,
    ) -> Tuple[List[Dict], Optional[str], Optional[datetime]]:
        """Get the latest datapoints."""
        query_string = (
            """
        query(
            $orgId: UUID!,
            $projectId: UUID!,
            $stageName: String,
            $cacheTime: DateTime,
            $first: Int,
            $cursor: String
        ) {
            tasksPaged(
                orgId: $orgId
                projectId: $projectId
                stageName: $stageName
                cacheTime: $cacheTime
                first: $first
                after: $cursor
            ) {
                entries {
                    taskId
                    dpId
                    currentStageName
                    %s
                }
                cursor
                cacheTime
            }
            }
        """
            % LATEST_TASKDATA_SHARD
        )
        # EXECUTE THE QUERY
        query_variables = {
            "orgId": org_id,
            "projectId": project_id,
            "stageName": stage_name,
            "cacheTime": None if cache_time is None else cache_time.isoformat(),
            "first": first,
            "cursor": cursor,
        }

        result = self.client.execute_query(query_string, query_variables, False)
        tasks_paged = result.get("tasksPaged", {}) or {}
        entries: List[Dict] = tasks_paged.get("entries", []) or []  # type: ignore
        new_cache_time: Optional[str] = tasks_paged.get("cacheTime")
        return (
            entries,
            tasks_paged.get("cursor"),
            parser.parse(new_cache_time) if new_cache_time else None,
        )

    def get_datapoint_latest(self, org_id: str, project_id: str, task_id: str) -> Dict:
        """Get the latest labels for a single bdatapoint."""
        query_string = (
            """
        query($orgId: UUID!, $projectId: UUID!, $taskId: UUID!) {
            task(
                orgId: $orgId
                projectId: $projectId
                taskId: $taskId
            ) {
                taskId
                currentStageName
                %s
            }
            }
        """
            % LATEST_TASKDATA_SHARD
        )
        # EXECUTE THE QUERY
        query_variables = {
            "orgId": org_id,
            "projectId": project_id,
            "taskId": task_id,
        }

        result: Dict[str, Dict] = self.client.execute_query(
            query_string, query_variables, False
        )

        return result.get("task", {}) or {}

    async def get_labels(
        self, session: aiohttp.ClientSession, org_id: str, project_id: str, dp_id: str
    ) -> Dict:
        """Get input labels."""
        query = """
        query dataPoint(
            $orgId: UUID!
            $dpId: UUID!
            $name: String!
        ) {
            dataPoint(orgId: $orgId, dpId: $dpId) {
                labelData(customGroupName: $name) {
                    dpId
                    labelsData
                }
            }
        }
        """

        variables = {
            "orgId": org_id,
            "dpId": dp_id,
            "name": f"{project_id}-input",
        }

        response = await self.client.execute_query_async(session, query, variables)
        label_data: Dict = response.get("dataPoint", {}).get("labelData", {})
        return label_data



"""Testing functions for Export."""


import sys
from typing import Dict
import copy

import numpy as np
from redbrick.export.public import Export

taxonomy_1 = {
    "categories": [
        {
            "name": "object",
            "children": [
                {
                    "name": "bus",
                    "classId": 0,
                    "children": [
                        {"name": "motor", "classId": 6, "children": []},
                    ],
                },
                {"name": "traffic light", "classId": 1, "children": []},
                {"name": "traffic sign", "classId": 2, "children": []},
                {
                    "name": "person",
                    "classId": 3,
                    "children": [
                        {"name": "bike", "classId": 4, "children": []},
                        {"name": "truck", "classId": 5, "children": []},
                    ],
                },
            ],
        }
    ],
}
tax_map_1 = {
    "bus": 1,
    "traffic light": 2,
    "traffic sign": 3,
    "person": 4,
    "bike": 5,
    "truck": 6,
    "motor": 7,
}


def test_tax_class_id_map() -> None:
    """Test taxonomy classid map."""
    tax_map: Dict = {}
    tax = taxonomy_1["categories"][0]["children"]
    Export.tax_class_id_mapping(tax, tax_map)  # type: ignore
    assert tax_map == tax_map_1


def test_png_convert_simple() -> None:
    """Test converting to a simple png."""
    if sys.platform not in ("linux", "darwin"):
        # Don't run this test for windows, w/o rasterio will fail.
        assert True
        return

    taxonomy = {
        "categories": [
            {
                "name": "object",
                "children": [{"name": "bus", "classId": 0, "children": []}],
            }
        ]
    }
    class_id_map: Dict = {}
    color_map: Dict = {}
    Export.tax_class_id_mapping(
        taxonomy["categories"][0]["children"], class_id_map, color_map  # type: ignore
    )

    task = {
        "labels": [
            {
                "category": [["object", "bus"]],
                "pixel": {
                    "imagesize": [10, 10],
                    "regions": [[[1, 1], [8, 1], [8, 8], [1, 8]]],
                    "holes": None,
                },
            }
        ]
    }
    color_mask = Export.convert_rbai_mask(task["labels"], class_id_map, False, 30)
    assert (color_mask[1:8, 1:8, :] / color_map["bus"] == np.ones((7, 7, 3))).all()


def test_fill_holes_simple() -> None:
    """Tests the fill holes operation."""
    mask = np.zeros((10, 10))
    mask[2:7, 2:7] = 1
    mask[3:6, 3:6] = 0

    mask_true = np.zeros((10, 10))
    mask_true[2:7, 2:7] = 1
    mask_filled = Export.fill_mask_holes(mask, 30)

    assert (mask_filled == mask_true).all()

    mask_filled_2 = Export.fill_mask_holes(mask, 4)
    assert (mask_filled_2 == mask).all()


def test_fill_holes_big_and_small() -> None:
    """Test the fill holes operation, for big and small holes."""
    mask = np.ones((50, 50))
    mask[20:40, 20:40] = 0
    mask_true = copy.deepcopy(mask)
    mask[2:7, 2:7] = 0

    mask_filled = Export.fill_mask_holes(mask, 26)
    assert (mask_filled == mask_true).all()



"""Async utils."""
import asyncio
from typing import Awaitable, List, TypeVar, Optional, Iterable
import tqdm.asyncio

from redbrick.common.constants import MAX_CONCURRENCY

ReturnType = TypeVar("ReturnType")


async def gather_with_concurrency(
    max_concurrency: int,
    tasks: Iterable[Awaitable[ReturnType]],
    progress_bar_name: Optional[str] = None,
    keep_progress_bar: bool = True,
) -> List[ReturnType]:
    """Utilizes a Semaphore to limit concurrency to n."""
    if not tasks:
        return []

    max_concurrency = min(max_concurrency, MAX_CONCURRENCY)
    semaphore = asyncio.Semaphore(max_concurrency)

    async def sem_task(task: Awaitable[ReturnType]) -> ReturnType:
        async with semaphore:
            return await task

    coros = [sem_task(task) for task in tasks]

    if progress_bar_name:
        result = []
        for coro in tqdm.asyncio.tqdm.as_completed(
            coros, desc=progress_bar_name, leave=keep_progress_bar
        ):
            temp = await coro
            result.append(temp)
        return result

    return await asyncio.gather(*coros)



"""Main CLI project."""
import os
from typing import Optional

from halo.halo import Halo  # type: ignore

from redbrick import _populate_context
from redbrick.common.context import RBContext
from redbrick.organization import RBOrganization
from redbrick.project import RBProject
from redbrick.cli.entity import CLICache, CLIConfiguration, CLICredentials
from redbrick.utils.logging import print_info


class CLIProject:
    """CLIProject class."""

    path: str

    creds: CLICredentials
    conf: CLIConfiguration
    cache: CLICache

    _context: Optional[RBContext] = None
    _org: Optional[RBOrganization] = None
    _project: Optional[RBProject] = None

    def __init__(self, path: str = ".", required: bool = True) -> None:
        """Initialize CLIProject."""
        self.path = os.path.realpath(path)
        assert os.path.isdir(self.path), f"Not a valid directory {self.path}"

        self._rb_dir = os.path.join(self.path, ".redbrick")
        self._creds_file = os.path.join(
            os.path.expanduser("~"), ".redbrickai", "credentials"
        )

        self.creds = CLICredentials(
            os.path.join(os.path.expanduser("~"), ".redbrickai", "credentials")
        )
        self.conf = CLIConfiguration(os.path.join(self._rb_dir, "config"))
        self.cache = CLICache(os.path.join(self._rb_dir, "cache"), self.conf)

        if required:
            assert (
                self.creds.exists
            ), "No credentials found, please set it up with `redbrick config`"
            assert self.conf.exists, (
                "No project found in {self.path}\n"
                + "Please create one using `redbrick init` / clone existing using `redbrick clone`"
            )
            assert (
                self.org_id == self.creds.org_id
            ), "Project configuration does not match with current profile"

    @classmethod
    def from_path(
        cls, path: str = ".", required: bool = True
    ) -> Optional["CLIProject"]:
        """Get CLIProject from given directory."""
        path = os.path.realpath(path)

        if os.path.isdir(os.path.join(path, ".redbrick")):
            return cls(path, required)

        parent = os.path.realpath(os.path.join(path, ".."))

        if parent == path:
            if required:
                raise Exception(f"No redbrick project found. Searched upto {path}")
            return None

        return cls.from_path(parent, required)

    @property
    def context(self) -> RBContext:
        """Get RedBrick context."""
        if not self._context:
            self._context = _populate_context(self.creds.context)
        return self._context

    @property
    def org_id(self) -> str:
        """Get org_id of current project."""
        value = self.conf.get_option("org", "id")
        assert value, "Invalid project configuration"
        return value.strip().lower()

    @property
    def project_id(self) -> str:
        """Get project_id of current project."""
        value = self.conf.get_option("project", "id")
        assert value, "Invalid project configuration"
        return value.strip().lower()

    @property
    def org(self) -> RBOrganization:
        """Get org object."""
        if not self._org:
            org = self.cache.get_object("org")
            if (
                isinstance(org, RBOrganization)
                and org.context.client.url == self.context.client.url
                and org.context.client.api_key == self.context.client.api_key
            ):
                self._org = org
            else:
                with Halo(text="Fetching organization", spinner="dots") as spinner:
                    try:
                        self._org = RBOrganization(self.context, self.org_id)
                        self.cache.set_object("org", self._org, True)
                    except Exception as error:
                        spinner.fail()
                        raise error
                spinner.succeed(str(self._org))
        return self._org

    @property
    def project(self) -> RBProject:
        """Get project object."""
        if not self._project:
            project = self.cache.get_object("project")
            if (
                isinstance(project, RBProject)
                and project.context.client.url == self.context.client.url
                and project.context.client.api_key == self.context.client.api_key
            ):
                self._project = project
            else:
                with Halo(text="Fetching project", spinner="dots") as spinner:
                    try:
                        self._project = RBProject(
                            self.context, self.org_id, self.project_id
                        )
                        self.cache.set_object("project", self._project, True)
                    except Exception as error:
                        spinner.fail()
                        raise error
                    spinner.succeed(str(self._project))
        return self._project

    def initialize_project(self, org: RBOrganization, project: RBProject) -> None:
        """Initialize local project."""
        assert not os.path.isdir(
            self._rb_dir
        ), f"Already a RedBrick project {self.path}"

        os.makedirs(self._rb_dir)
        self.conf.save()

        self.conf.set_section("org", {"id": project.org_id})
        self.conf.set_section("project", {"id": project.project_id})

        self.cache.set_object("org", org)
        self.cache.set_object("project", project)

        self.conf.save()

        print_info(
            f"Successfully initialized {project} in {self.path}\nURL: {project.url}"
        )



"""CLI clone command."""
import os
import re
from argparse import ArgumentParser, Namespace

from halo.halo import Halo  # type: ignore

from redbrick.cli.input.select import CLIInputSelect
from redbrick.cli.project import CLIProject
from redbrick.cli.cli_base import CLICloneInterface
from redbrick.organization import RBOrganization
from redbrick.project import RBProject


class CLICloneController(CLICloneInterface):
    """CLI clone command controller."""

    def __init__(self, parser: ArgumentParser) -> None:
        """Intialize clone sub commands."""
        parser.add_argument("project", nargs="?", help="Project ID or Name")
        parser.add_argument("path", nargs="?", help="Local path of project")

    def handler(self, args: Namespace) -> None:
        """Handle clone command."""
        self.args = args

        if self.args.path is not None:
            path = os.path.realpath(self.args.path)
            assert not os.path.exists(path), f"{path} already exists"

        project = CLIProject.from_path(
            path="." if self.args.path is None else self.args.path, required=False
        )
        assert project is None, f"Already in a project {project.path}"

        self.handle_clone()

    def handle_clone(self) -> None:
        """Handle empty sub command."""
        temp = CLIProject(required=False)
        assert temp.creds.exists, "Credentials missing"

        with Halo(text="Fetching organization", spinner="dots") as spinner:
            try:
                org = RBOrganization(temp.context, temp.creds.org_id)
                spinner.succeed(str(org))
            except Exception as error:
                spinner.fail()
                raise error

        with Halo(text="Fetching projects", spinner="dots") as spinner:
            try:
                projects = temp.context.project.get_projects(org.org_id)
                spinner.succeed()
            except Exception as error:
                spinner.fail()
                raise error

        projects = list(
            filter(lambda proj: proj["status"] == "CREATION_SUCCESS", projects)
        )

        project_ids, project_names, choices = [], [], []
        for proj in projects:
            project_ids.append(proj["projectId"])
            project_names.append(proj["name"])
            choices.append(proj["name"] + " (" + proj["projectId"] + ")")

        selected = self.args.project
        if selected:
            index = -1
            if selected in project_ids:
                index = project_ids.index(selected)
            elif selected in project_names:
                index = project_names.index(selected)

            if index >= 0:
                selected = project_names[index] + " (" + project_ids[index] + ")"

        selected_index = choices.index(
            CLIInputSelect(selected, "Project", choices).get()
        )

        if self.args.path is not None:
            path = os.path.realpath(self.args.path)
        else:
            path = os.path.join(
                os.path.realpath("."),
                re.sub(r"\W+", "-", project_names[selected_index]),
            )

        os.makedirs(path)

        project = CLIProject(path=path, required=False)
        project.initialize_project(
            org, RBProject(org.context, org.org_id, project_ids[selected_index])
        )



"""Test create project, upload labels and export data."""
import contextlib
from datetime import datetime
import random
from uuid import uuid4
from typing import Dict, Generator, List, Optional, Tuple
import time

import pytest
import tenacity

import redbrick
from redbrick.common.enums import LabelType, StorageMethod
from redbrick.project import RBProject


@contextlib.contextmanager
def create_test_project(
    org_id: str,
    api_key: str,
    url: str,
    label_type: LabelType,
    taxonomy: str = "DEFAULT::Berkeley Deep Drive (BDD)",
    reviews: int = 1,
) -> Generator[RBProject, None, None]:
    """Create project."""
    project_name = (
        f"{label_type.value}"
        + f"{datetime.strftime(datetime.now(), '%Y%m%d%H%M%S%f')}{random.randint(0, 1000)}"
    )
    project: Optional[RBProject] = None
    try:
        org = redbrick.get_org(api_key=api_key, org_id=org_id, url=url)
        project = org.create_project(
            name=project_name,
            label_type=label_type,
            taxonomy_name=taxonomy,
            reviews=reviews,
        )
        assert project
        yield project
    finally:
        if project:
            project.context.project.delete_project(project.org_id, project.project_id)


def upload_test_data(project: RBProject, num_tasks: int) -> None:
    """Upload test data."""
    tasks = [
        {
            "name": str(uuid4()),
            "items": ["https://datasets.redbrickai.com/bccd/BloodImage_00000.jpg"],
        }
        for _ in range(num_tasks)
    ]
    project.upload.create_datapoints(
        StorageMethod.PUBLIC,
        tasks,
    )


def get_test_label_classify() -> List[Dict]:
    """Get classify test label."""
    categories = ["bus", "bike", "truck", "motor", "car", "train", "rider"]
    return [
        {
            "labelid": str(uuid4()),
            "category": [["object", random.choice(categories)]],
            "attributes": [],
            "taskclassify": True,
        }
    ]


def get_test_label_polygon() -> List[Dict]:
    """Get polygon test label."""
    categories = ["bus", "bike", "truck", "motor", "car", "train", "rider"]
    return (
        [
            {
                "labelid": str(uuid4()),
                "category": [["object", random.choice(categories)]],
                "attributes": [],
                "polygon": [
                    {
                        "xnorm": random.random(),
                        "ynorm": random.random(),
                    }
                    for _ in range(random.randint(3, 10))
                ],
            }
            for _ in range(random.randint(1, 3))
        ]
        if random.random() > 0.5
        else []
    )


def label_test_data(
    project: RBProject, num_tasks: int
) -> Tuple[List[Dict], List[Dict]]:
    """Label test data."""
    all_tasks, labeled_tasks = [], []
    if project.project_type == LabelType.IMAGE_CLASSIFY:
        labels = get_test_label_classify
    elif project.project_type == LabelType.IMAGE_POLYGON:
        labels = get_test_label_polygon

    tasks: List[Dict] = []
    for attempt in tenacity.Retrying(
        reraise=True,
        stop=tenacity.stop_after_attempt(10),
        wait=tenacity.wait_exponential(multiplier=1, min=1, max=10),
        retry=tenacity.retry_if_not_exception_type(
            (KeyboardInterrupt, PermissionError, ValueError)
        ),
    ):
        with attempt:
            tasks = project.labeling.get_tasks("Label", num_tasks)
            assert len(tasks) == num_tasks

    to_label = random.randint(1, num_tasks)
    cur_batch = [
        {
            **task,
            "labels": labels(),
        }
        for task in tasks[:to_label]
    ]
    labeled_tasks.extend(cur_batch)
    all_tasks.extend(
        cur_batch
        + [
            {
                **task,
                "labels": [],
            }
            for task in tasks[to_label:]
        ]
    )
    if cur_batch:
        project.labeling.put_tasks(
            "Label",
            [
                {"taskId": task["taskId"], "labels": task["labels"]}
                for task in cur_batch
            ],
        )

    return all_tasks, labeled_tasks


def review_test_data(project: RBProject, labeled_tasks: List[Dict]) -> List[Dict]:
    """Review test data."""
    reviewed_tasks = []

    for attempt in tenacity.Retrying(
        reraise=True,
        stop=tenacity.stop_after_attempt(10),
        wait=tenacity.wait_exponential(multiplier=1, min=1, max=10),
        retry=tenacity.retry_if_not_exception_type(
            (KeyboardInterrupt, PermissionError, ValueError)
        ),
    ):
        with attempt:
            tasks = project.review.get_tasks(
                "Review_1", random.randint(1, len(labeled_tasks))
            )
            assert tasks

    cur_batch = [
        {
            "taskId": task["taskId"],
            "reviewVal": random.random() > 0.5,
        }
        for task in tasks
    ]
    reviewed_tasks.extend(cur_batch)
    if cur_batch:
        project.review.put_tasks("Review_1", cur_batch)

    return reviewed_tasks


def validate_redbrick_labels(tasks: Dict[str, Dict], exported: List[Dict]) -> None:
    """Validate redbrick labels."""
    for task in exported:
        assert (
            task["taskId"] in tasks
            and tasks[task["taskId"]]["labels"] == task["labels"]
        )


def validate_export_data_redbrick(
    project: RBProject, tasks: Dict[str, Dict], groundtruth_task_ids: List[str]
) -> None:
    """Export and validate test data for redbrick format."""
    # All tasks
    exported_data = project.export.redbrick_format(False)
    assert set({task["taskId"] for task in exported_data}) == set(tasks.keys())
    validate_redbrick_labels(tasks, exported_data)

    # Groundtruth
    exported_data = project.export.redbrick_format()
    assert set({task["taskId"] for task in exported_data}) == set(groundtruth_task_ids)
    validate_redbrick_labels(tasks, exported_data)

    # TaskId
    task_id = random.choice(list(tasks.keys()))
    exported_data = project.export.redbrick_format(task_id=task_id)
    assert set({task["taskId"] for task in exported_data}) == set([task_id])
    validate_redbrick_labels(tasks, exported_data)


# pylint: disable=unused-argument
def validate_coco_labels(tasks: Dict[str, Dict], exported: Dict) -> None:
    """Validate coco labels."""


def validate_export_data_coco(
    project: RBProject, tasks: Dict[str, Dict], groundtruth_task_ids: List[str]
) -> None:
    """Export and validate test data for coco format."""
    # All tasks
    exported_data = project.export.coco_format(False)
    assert len(exported_data["images"]) == len(tasks)
    validate_coco_labels(tasks, exported_data)

    # Groundtruth
    exported_data = project.export.coco_format()
    assert len(exported_data["images"]) == len(groundtruth_task_ids)
    validate_coco_labels(tasks, exported_data)

    # TaskId
    task_id = random.choice(list(tasks.keys()))
    exported_data = project.export.coco_format(task_id=task_id)
    assert len(exported_data["images"]) == 1
    validate_coco_labels(tasks, exported_data)


@pytest.mark.slow
@pytest.mark.parametrize(
    "label_type", (LabelType.IMAGE_CLASSIFY, LabelType.IMAGE_POLYGON)
)
def test_classify_project(
    org_id: str, api_key: str, url: str, label_type: LabelType
) -> None:
    """Test export."""
    with create_test_project(org_id, api_key, url, label_type) as project:
        task_count = 20
        upload_test_data(project, task_count)
        all_tasks, labeled_tasks = label_test_data(project, task_count)
        reviewed_tasks = review_test_data(project, labeled_tasks)

        tasks_map = {task["taskId"]: task for task in all_tasks}
        groundtruth_task_ids = [
            task["taskId"] for task in reviewed_tasks if task["reviewVal"]
        ]

        time.sleep(10)  # give time for tasks to travel across the queues

        validate_export_data_redbrick(project, tasks_map, groundtruth_task_ids)
        if label_type in (LabelType.IMAGE_POLYGON,):
            validate_export_data_coco(project, tasks_map, groundtruth_task_ids)



"""
RedBrick SDK is a Python package for the RedBrick AI platform.

Visit https://docs.redbrickai.com/python-sdk/sdk-overview for an
overview of how to use the SDK and to view code examples.

To use the RedBrick SDK you need to create an API key. Please
see this documentation for accomplishing that.
https://docs.redbrickai.com/python-sdk/sdk-overview#generate-an-api-key
"""
import sys
import asyncio

from redbrick.common.context import RBContext
from redbrick.common.enums import LabelType, StorageMethod
from redbrick.common.constants import (
    DEFAULT_URL,
    ORG_API_HAS_CHANGED,
    PROJECT_API_HAS_CHANGED,
)
from redbrick.project import RBProject
from redbrick.organization import RBOrganization

from redbrick.utils.logging import print_info, handle_exception

from .version_check import version_check

__version__ = "1.1.2"

# windows event loop close bug https://github.com/encode/httpx/issues/914#issuecomment-622586610
try:
    if (
        sys.version_info[0] == 3
        and sys.version_info[1] >= 8
        and sys.platform.startswith("win")
    ):
        asyncio.set_event_loop_policy(  # type: ignore
            asyncio.WindowsSelectorEventLoopPolicy()  # pylint: disable=no-member
        )
except Exception:  # pylint: disable=broad-except
    pass


version_check(__version__)


def _populate_context(context: RBContext) -> RBContext:
    # pylint: disable=import-outside-toplevel
    from redbrick.repo import (
        ExportRepo,
        LabelingRepo,
        LearningRepo,
        Learning2Repo,
        UploadRepo,
        ProjectRepo,
    )

    context.export = ExportRepo(context.client)
    context.labeling = LabelingRepo(context.client)
    context.learning = LearningRepo(context.client)
    context.learning2 = Learning2Repo(context.client)
    context.upload = UploadRepo(context.client)
    context.project = ProjectRepo(context.client)
    return context


@handle_exception
def get_org(org_id: str, api_key: str, url: str = DEFAULT_URL) -> RBOrganization:
    """
    Get an existing redbrick organization object.

    Organization object allows you to interact with your organization
    and perform high level actions like creating a project.

    Parameters
    ---------------
    org_id: str
        Your organizations unique id https://app.redbrickai.com/<org_id>/.

    api_key: str
        Your secret api_key, can be created from the RedBrick AI platform.

    url: str = DEFAULT_URL
        Should default to https://api.redbrickai.com
    """
    if len(org_id) != 36:
        raise ValueError("Your first argument looks incorrect, " + ORG_API_HAS_CHANGED)
    if "." in api_key:
        raise ValueError(
            "Your second argument looks like a url, " + ORG_API_HAS_CHANGED
        )

    context = _populate_context(RBContext(api_key=api_key, url=url))

    return RBOrganization(context, org_id)


@handle_exception
def get_project(
    org_id: str, project_id: str, api_key: str, url: str = DEFAULT_URL
) -> RBProject:
    """
    Get an existing RedBrick project object.

    Project objects allow you to interact with your RedBrick Ai projects,
    and perform actions like importing data, exporting data etc.

    Parameters
    ---------------
    org_id: str
        Your organizations unique id https://app.redbrickai.com/<org_id>/

    project_id: str
        Your projects unique id https://app.redbrickai.com/<org_id>/<project_id>/

    api_key: str
        Your secret api_key, can be created from the RedBrick AI platform.

    url: str = DEFAULT_URL
        Should default to https://api.redbrickai.com
    """
    if len(org_id) != 36:
        raise ValueError(
            "Your first argument looks incorrect, " + PROJECT_API_HAS_CHANGED
        )
    if "http" in project_id:
        raise ValueError(
            "Your second argument looks like a url, " + PROJECT_API_HAS_CHANGED
        )

    context = _populate_context(RBContext(api_key=api_key, url=url))
    return RBProject(context, org_id, project_id)



"""Public interface to upload module."""
import uuid
import asyncio
import os
from copy import deepcopy
from typing import List, Dict, Optional
import json
from uuid import uuid4

import aiohttp
import numpy as np
from shapely.geometry import MultiPolygon, shape as gen_shape

from redbrick.common.context import RBContext
from redbrick.common.enums import LabelType
from redbrick.common.constants import MAX_CONCURRENCY
from redbrick.common.enums import StorageMethod
from redbrick.utils.async_utils import gather_with_concurrency
from redbrick.utils.logging import print_error, handle_exception
from redbrick.utils.segmentation import check_mask_map_format
from redbrick.utils.files import (
    NIFTI_FILE_TYPES,
    VIDEO_FILE_TYPES,
    get_file_type,
    upload_files,
)


class Upload:
    """Primary interface to uploading new data to a project."""

    def __init__(
        self, context: RBContext, org_id: str, project_id: str, project_type: LabelType
    ) -> None:
        """Construct Upload object."""
        self.context = context
        self.org_id = org_id
        self.project_id = project_id
        self.project_type = project_type

    async def _create_datapoint(
        self,
        session: aiohttp.ClientSession,
        storage_id: str,
        point: Dict,
        is_ground_truth: bool,
        labels_path: Optional[str] = None,
    ) -> Dict:
        """Try to create a datapoint."""
        try:
            # Basic structural validations, rest handled by API
            assert (
                isinstance(point, dict) and point
            ), "Task object must be a non-empty dictionary"
            assert (
                "response" not in point and "error" not in point
            ), "Task object must not contain `response` or `error`"
            assert (
                "name" in point and isinstance(point["name"], str) and point["name"]
            ), "Task object must contain a valid `name`"
            assert (
                "items" in point
                and isinstance(point["items"], list)
                and (
                    len(point["items"]) == 1
                    if str(self.project_type.value).startswith("IMAGE_")
                    else point["items"]
                )
                and all(
                    map(lambda item: isinstance(item, str) and item, point["items"])
                )
            ), "`items` must be a list of urls (one for image and multiple for videoframes)"
            assert "labels" not in point or (
                isinstance(point["labels"], list)
                and all(
                    map(
                        lambda label: isinstance(label, dict) and label, point["labels"]
                    )
                )
            ), "`labels` must be a list of label objects"
            response = await self.context.upload.create_datapoint_async(
                session,
                self.org_id,
                self.project_id,
                storage_id,
                point["name"],
                point["items"],
                json.dumps(point.get("labels", [])),
                labels_path,
                is_ground_truth,
            )
            assert response.get("taskId"), "Failed to create task"
            point_success = deepcopy(point)
            point_success["response"] = response
            return point_success
        except Exception as error:  # pylint:disable=broad-except
            print_error(error)
            point_error = deepcopy(point)
            point_error["error"] = error
            return point_error

    async def _create_datapoints(
        self, storage_id: str, points: List[Dict], is_ground_truth: bool
    ) -> List[Dict]:
        conn = aiohttp.TCPConnector(limit=MAX_CONCURRENCY)
        async with aiohttp.ClientSession(connector=conn) as session:
            coros = [
                self._create_datapoint(session, storage_id, point, is_ground_truth)
                for point in points
            ]
            tasks = await gather_with_concurrency(
                MAX_CONCURRENCY, coros, "Creating datapoints"
            )

        await asyncio.sleep(0.250)  # give time to close ssl connections
        return tasks

    async def _item_upload(
        self,
        session: aiohttp.ClientSession,
        storage_id: str,
        file_name: str,
        file_path: str,
        is_ground_truth: bool,
    ) -> Dict:
        """Try to upload an item."""
        try:
            data_type, task_type = str(self.project_type.value).split("_", 1)
            response = await self.context.upload.item_upload_async(
                session,
                self.org_id,
                self.project_id,
                storage_id,
                data_type,
                task_type,
                file_path,
                file_name,
                get_file_type(file_path)[1],
                is_ground_truth,
            )
            assert response.get("ok"), "Failed to create task"
            return {"name": file_name, "filePath": file_path, "response": response}
        except Exception as error:  # pylint:disable=broad-except
            print_error(error)
            return {"name": file_name, "filePath": file_path, "error": error}

    async def _create_task(
        self,
        session: aiohttp.ClientSession,
        storage_id: str,
        point: Dict,
        is_ground_truth: bool,
    ) -> Dict:
        if storage_id == StorageMethod.REDBRICK:
            file_types, upload_items, presigned_items = [], [], []
            try:
                for item in point["items"]:
                    file_types.append(get_file_type(item)[1])
                    upload_items.append(os.path.split(item)[-1])
                presigned_items = self._generate_upload_presigned_url(
                    upload_items, file_types
                )
            except Exception:  # pylint:disable=broad-except
                print_error(f"Failed to upload {point['name']}")
                return {
                    "name": point["name"],
                    "error": f"Failed to upload {point['name']}",
                }

            files = [
                (
                    item,
                    presigned_items[idx]["presignedUrl"],
                    file_types[idx],
                )
                for idx, item in enumerate(point["items"])
            ]
            uploaded = await upload_files(
                files,
                f"Uploading items for {point['name'][:57]}{point['name'][57:] and '...'}"
                if len(point["items"]) > 1
                else None,
                False,
            )

            if not all(uploaded):
                print_error(f"Failed to upload {point['name']}")
                return {
                    "name": point["name"],
                    "error": f"Failed to upload {point['name']}",
                }

            point["items"] = [
                presigned_items[idx]["filePath"] for idx in range(len(point["items"]))
            ]

        labels_path: Optional[str] = None
        data_type = str(self.project_type.value).split("_", 1)[0]
        if (
            data_type == "DICOM"
            and point.get("labelsPath")
            and str(point["labelsPath"]).endswith(".nii")
            and os.path.isfile(point["labelsPath"])
        ):
            file_type = NIFTI_FILE_TYPES["nii"]
            presigned = await self.context.labeling.presign_labels_path(
                session, self.org_id, self.project_id, str(uuid4()), file_type
            )
            if (
                await upload_files(
                    [(point["labelsPath"], presigned["presignedUrl"], file_type)],
                    f"Uploading labels for {point['name'][:57]}{point['name'][57:] and '...'}",
                    False,
                )
            )[0]:
                labels_path = presigned["filePath"]

        elif (
            data_type == "VIDEO"
            and get_file_type(point["items"][0])[0] in VIDEO_FILE_TYPES
        ):
            return await self._item_upload(
                session,
                storage_id,
                point["name"],
                point["items"][0],
                is_ground_truth,
            )

        return await self._create_datapoint(
            session, storage_id, point, is_ground_truth, labels_path
        )

    async def _create_tasks(
        self,
        storage_id: str,
        points: List[Dict],
        is_ground_truth: bool,
    ) -> List[Dict]:
        conn = aiohttp.TCPConnector(limit=MAX_CONCURRENCY)
        async with aiohttp.ClientSession(connector=conn) as session:
            coros = [
                self._create_task(session, storage_id, point, is_ground_truth)
                for point in points
            ]
            tasks = await gather_with_concurrency(
                MAX_CONCURRENCY, coros, "Creating tasks"
            )

        await asyncio.sleep(0.250)  # give time to close ssl connections
        return tasks

    def _generate_upload_presigned_url(
        self, files: List[str], file_type: List[str]
    ) -> List[Dict]:
        """
        Generate presigned url's to perform upload.

        Parameters:
        -------------
        files: List[str]
            This needs to be names of the files when it's uploaded.
            i.e. locally, the file can be image.png, but after
            uploading if you want it named image001.png, the files
            List must contain [image001.png].

        file_type: List[str]
            Corresponding file types.

        Returns
        ------------
        List[Dict]
            [
                {
                    "presignedUrl: "...", # url to upload to
                    "filePath": "..." # remote file path
                    "fileName": "..." # the same as values in files
                }
            ]
        """
        try:
            result = self.context.upload.items_upload_presign(
                self.org_id, self.project_id, files, file_type
            )
        except ValueError as error:
            print_error(error)
            raise error
        return result

    @staticmethod
    def _check_validity_of_items(items: Optional[List[str]]) -> List[str]:
        """
        Check the validity of an items entry with locally stored data.

        Returns
        ---------
        List[str]
            List of invalid items.
        """
        invalid = []
        if not (isinstance(items, list) and items):
            raise ValueError(
                "`items` must be a list of urls (one for image and multiple for videoframes)"
            )
        for item in items:
            if not (isinstance(item, str) and item and os.path.isfile(item)):
                invalid += [item]
                print_error(f"{item} is an invalid path")
                continue
            try:
                get_file_type(item)
            except ValueError as error:
                invalid += [item]
                print_error(error)

        return invalid

    @staticmethod
    def _mask_to_polygon(
        mask: np.ndarray,
    ) -> MultiPolygon:
        """Convert masks to polygons."""
        try:
            import rasterio  # pylint: disable=import-error, import-outside-toplevel
            from rasterio import (  # pylint: disable=import-error, import-outside-toplevel
                features,
            )
        except Exception as error:
            print_error(
                "For windows users, please follow the rasterio "
                + "documentation to properly install the module "
                + "https://rasterio.readthedocs.io/en/latest/installation.html "
                + "Rasterio is required by RedBrick SDK to work with masks."
            )
            raise error

        all_polygons = []
        for shape, _ in features.shapes(
            mask.astype(np.int16),
            mask=(mask > 0),
            transform=rasterio.Affine(1.0, 0, 0, 0, 1.0, 0),
        ):
            all_polygons.append(gen_shape(shape))

        polygon = MultiPolygon(all_polygons)
        if not polygon.is_valid:
            polygon = polygon.buffer(0)
            # Sometimes buffer() converts a simple Multipolygon to just a Polygon,
            # need to keep it a Multi throughout
            if polygon.type == "Polygon":
                polygon = MultiPolygon([polygon])
        return polygon

    @staticmethod
    def mask_to_rbai(  # pylint: disable=too-many-locals
        mask: np.ndarray, class_map: Dict, items: str, name: str
    ) -> Dict:
        """Convert a mask to rbai datapoint format."""
        # Convert 3D mask into a series of 2D masks for each object
        mask_2d_stack: np.ndarray = np.array([])
        mask_2d_categories = []
        for _, category in enumerate(class_map):
            mask_2d = np.zeros((mask.shape[0], mask.shape[1]))
            color = class_map[category]
            class_idxs = np.where(
                (mask[:, :, 0] == color[0])
                & (mask[:, :, 1] == color[1])
                & (mask[:, :, 2] == color[2])
            )

            if len(class_idxs[0]) == 0:
                # Skip classes that aren't present
                continue

            mask_2d[class_idxs] = 1  # fill in binary mask
            mask_2d_categories += [category]

            # Stack all individual masks
            if mask_2d_stack.shape[0] == 0:
                mask_2d_stack = mask_2d
            else:
                mask_2d_stack = np.dstack((mask_2d_stack, mask_2d))  # type: ignore

        entry: Dict = {}
        entry["labels"] = []
        for depth in range(mask_2d_stack.shape[-1]):
            mask_depth = mask_2d_stack[:, :, depth]
            polygons = Upload._mask_to_polygon(mask_depth)

            label_entry: Dict = {}
            label_entry["category"] = [["object", mask_2d_categories[depth]]]
            label_entry["attributes"] = []
            label_entry["pixel"] = {}
            label_entry["pixel"]["imagesize"] = [
                mask_depth.shape[1],
                mask_depth.shape[0],
            ]
            label_entry["pixel"]["regions"] = []
            label_entry["pixel"]["holes"] = []
            for polygon in polygons.geoms:
                label_entry["pixel"]["regions"].append(
                    [list(map(int, coords)) for coords in polygon.exterior.coords]
                )
                for interior in polygon.interiors:
                    label_entry["pixel"]["holes"].append(
                        [list(map(int, coords)) for coords in interior.coords]
                    )

            entry["labels"] += [label_entry]

        entry["items"] = [items]
        entry["name"] = name

        return entry

    @handle_exception
    def create_datapoints(
        self,
        storage_id: str,
        points: List[Dict],
        is_ground_truth: bool = False,
    ) -> List[Dict]:
        """
        Create datapoints in project.

        Optionally you can upload labels with your data.

        >>> project = redbrick.get_project(api_key, url, org_id, project_id)
        >>> project.upload.create_datapoints(...)

        Parameters
        --------------
        storage_id: str
            Your RedBrick AI external storage_id. This can be found under the Storage Tab
            on the RedBrick AI platform. To directly upload images to rbai,
            use redbrick.StorageMathod.REDBRICK.

        points: List[Dict]
            Please see the RedBrick AI reference documentation for overview of the format.
            https://docs.redbrickai.com/python-sdk/importing-data-and-labels

        is_ground_truth: bool = False
            If labels are provided in points above, and this parameters is set to true, the labels
            will be added to the Ground Truth stage. This is mainly useful for Active Learning.

        Returns
        -------------
        List[Dict]
            List of task objects with key `response` if successful, else `error`

        Notes
        ----------
            1. If doing direct upload, please use ``redbrick.StorageMethod.REDBRICK``
            as the storage id. Your items path must be a valid path to a locally stored image.

            2. When doing direct upload i.e. ``redbrick.StorageMethod.REDBRICK``,
            if you didn't specify a "name" field in your datapoints object,
            we will assign the "items" path to it.

        """
        created, skipped, filtered_points = [], [], []
        # Check if user wants to do a direct upload
        if storage_id == StorageMethod.REDBRICK:
            for point in points:
                # check path of items, if invalid, skip
                invalid = Upload._check_validity_of_items(point.get("items"))
                if invalid:
                    point_error = deepcopy(point)
                    point_error["error"] = "Invalid items"
                    skipped.append(point_error)
                    continue
                # update points dict with correct information
                if "name" not in point:
                    point["name"] = point["items"][0]

                filtered_points.append(deepcopy(point))
        else:
            filtered_points = list(map(deepcopy, points))  # type: ignore

        if filtered_points:
            loop = asyncio.get_event_loop()
            created = loop.run_until_complete(
                self._create_tasks(storage_id, filtered_points, is_ground_truth)
            )

        return skipped + created

    @handle_exception
    def create_datapoint_from_masks(
        self,
        storage_id: str,
        mask: np.ndarray,
        class_map: Dict,
        image_path: str,
    ) -> Dict:
        """
        Create a single datapoint with mask.

        >>> project = redbrick.get_project(api_key, url, org_id, project_id)
        >>> project.upload.create_datapoint_from_masks(...)

        Parameters
        --------------
        storage_id: str
            Your RedBrick AI external storage_id. This can be found under the Storage Tab
            on the RedBrick AI platform. Currently, this method only supports external storage
            or public storage i.e. redbrick.StorageMathod.PUBLIC (public hosted data).

        mask: np.ndarray.astype(np.uint8)
            A RGB mask with values as np.uint8. The RGB values correspond to a category,
            who's mapping can be gound in class_map.

        class_map: Dict
            Maps between the category_name in your Project Taxonomy and the RGB color of the mask.

            >>> class_map
            >>> {'category-1': [12, 22, 93], 'category-2': [1, 23, 128]...}

        image_map: str
            A valid path to the corresponding image. If storage_id is
            redbrick.StorageMethod.REDBRICK, image_map must be a
            valid path to a locally stored image.

        Returns
        -------------
        Dict
            Task object with key `response` if successful, else `error`

        Warnings
        ------------
        The category names in class_map must be valid entries in your project taxonomy. The project
        type also must be IMAGE_SEGMENTATION.

        """
        check_mask_map_format(mask, class_map)

        loop = asyncio.get_event_loop()
        if storage_id == StorageMethod.REDBRICK:
            _, file_type = get_file_type(image_path)
            upload_filename = str(uuid.uuid4())

            # get pre-signed url for s3 upload
            presigned_items = self._generate_upload_presigned_url(
                [upload_filename], [file_type]
            )

            # upload to s3
            loop.run_until_complete(
                upload_files(
                    [
                        (
                            image_path,  # filename on local machine
                            presigned_items[0]["presignedUrl"],
                            file_type,
                        )
                    ],
                    None,
                )
            )

            # create datapoint
            items = presigned_items[0]["filePath"]
            name = image_path
            datapoint_entry = Upload.mask_to_rbai(mask, class_map, items, name)
            return loop.run_until_complete(
                self._create_datapoints(storage_id, [datapoint_entry], False)
            )[0]

        # If storage_id is for remote storage
        items = image_path
        name = image_path

        # create datapoint
        datapoint_entry = Upload.mask_to_rbai(mask, class_map, items, name)
        return loop.run_until_complete(
            self._create_datapoints(storage_id, [datapoint_entry], False)
        )[0]



"""Convert between rb and coco for bbox labels."""

from typing import Dict, List


def rb2coco_bbox(
    rb_label: Dict,
    label_id: int,
    image_id: int,
    category_id: int,
    width: int,
    height: int,
) -> Dict:
    """Convert rb bbox to coco bbox."""
    assert rb_label["bbox2d"]

    xnorm = rb_label["bbox2d"]["xnorm"]
    ynorm = rb_label["bbox2d"]["ynorm"]
    wnorm = rb_label["bbox2d"]["wnorm"]
    hnorm = rb_label["bbox2d"]["hnorm"]
    return {
        "id": label_id,
        "image_id": image_id,
        "category_id": category_id,
        "bbox": [
            int(xnorm * width),
            int(ynorm * height),
            int(wnorm * width),
            int(hnorm * height),
        ],
        "iscrowd": 0,
        "area": int(wnorm * width) * int(hnorm * height),
        "segmentation": [],
    }


def is_coco_bbox(coco_label: Dict) -> bool:
    """Check if given label is a valid "bbox" only label."""
    has_bbox = "bbox" in coco_label and len(coco_label["bbox"]) == 4
    no_segment = not coco_label.get("segmentation")
    no_keypoints = not coco_label.get("keypoints")
    return has_bbox and no_segment and no_keypoints


def coco2rb_bbox(
    coco_label: Dict, category: List[List[str]], width: int, height: int
) -> Dict:
    """Convert coco bbox to rb bbox."""
    assert is_coco_bbox(coco_label)

    xnorm = coco_label["bbox"][0] / width
    ynorm = coco_label["bbox"][1] / height
    wnorm = coco_label["bbox"][2] / width
    hnorm = coco_label["bbox"][3] / height

    xnorm = max(xnorm, 0)
    ynorm = max(ynorm, 0)
    if xnorm + wnorm > 1:
        wnorm = 1 - xnorm
    if ynorm + hnorm > 1:
        hnorm = 1 - ynorm

    return {
        "category": category,
        "bbox2d": {
            "xnorm": xnorm,
            "ynorm": ynorm,
            "wnorm": wnorm,
            "hnorm": hnorm,
        },
    }



"""Tests for upload module."""
import json
import sys
from typing import List, Dict

from uuid import uuid4
import matplotlib.pyplot as plt
import pytest

from redbrick.common.enums import StorageMethod
from redbrick.utils.files import get_file_type
from redbrick.upload.public import Upload
from redbrick.tests.conftest import *  # pylint: disable=wildcard-import, unused-wildcard-import


def test_mask_rbai() -> None:
    """Test mask to polygon conversion."""
    if sys.platform not in ("linux", "darwin"):
        # Don't run this test for windows, w/o rasterio will fail.
        assert True
        return
    # read mask that was generated by rbai export
    # i.e. rbai polygon -> png mask format
    mask = plt.imread(
        "redbrick/upload/mask_test/88664c8f-c6f8-4d5a-918e-41d8441a4509.png"
    )
    with open(
        "redbrick/upload/mask_test/class_map.json", "r", encoding="utf-8"
    ) as file:
        class_map = json.load(file)

    # convert the mask to rbai polygon
    dp_entry = Upload.mask_to_rbai(mask, class_map, "items123", "name123")

    # read original rbai polygon format
    with open(
        "redbrick/upload/mask_test/redbrick_export_segment.json", "r", encoding="utf-8"
    ) as file:
        datapoints = json.load(file)
    dp_entry_real = datapoints[0]

    # compare the two polygon formats to confirm mask -> polygon is correct
    for label_real in dp_entry_real["labels"]:
        for label in dp_entry["labels"]:
            if label_real["category"] == label["category"]:
                assert (
                    label_real["pixel"]["regions"].sort()
                    == label["pixel"]["regions"].sort()
                )

                if label_real["pixel"]["holes"]:
                    assert (
                        label_real["pixel"]["holes"].sort()
                        == label["pixel"]["holes"].sort()
                    )


def test_file_type_extraction() -> None:
    """Tests the extraction of MIME file type."""
    filepath = "folder/subfolder/image.png"
    filetype = get_file_type(filepath)
    assert filetype[0] == "png" and filetype[1] == "image/png"


def test_file_type_extraction_invalid() -> None:
    """Check invalid file extraction."""
    filepath = "folder/subfolder/image.csv"
    try:
        get_file_type(filepath)
        assert False
    except ValueError as error:
        assert type(error).__name__ == "ValueError"


def test_items_validity() -> None:
    """Check invalid items."""
    # pylint: disable=protected-access
    file_name = f"{uuid4()}.png"
    with open(file_name, "w", encoding="utf-8") as file_:
        file_.write("test")

    items = [2, "", "missing.txt", file_name]
    invalid = Upload._check_validity_of_items(items)  # type: ignore
    os.remove(file_name)
    assert len(invalid) == len(items) - 1


@pytest.mark.slow
def test_invalid_upload_object(project: RBProject) -> None:
    """Check invalid item list upload."""
    tasks: List[Dict] = [
        {},
        {"name": None},
        {"items": []},
        {"name": "test"},
        {"name": "test", "items": ""},
        {"name": "test", "items": []},
        {"name": "test", "items": [0]},
        {
            "name": "test",
            "items": [
                "https://datasets.redbrickai.com/bccd/BloodImage_00000.jpg",
                "https://datasets.redbrickai.com/bccd/BloodImage_00000.jpg",
            ],
        },
        {
            "name": "test",
            "items": ["https://datasets.redbrickai.com/bccd/BloodImage_00000.jpg"],
            "labels": {},
        },
        {
            "name": str(uuid4()),
            "items": ["https://datasets.redbrickai.com/bccd/BloodImage_00000.jpg"],
        },
    ]
    task_objects = project.upload.create_datapoints(StorageMethod.PUBLIC, tasks)
    assert (
        len(task_objects) == len(tasks)
        and len(list(filter(lambda task: "response" in task, task_objects))) == 1
    )
